create sequence "id" start 100 increment 1 maxvalue 2147483647 minvalue 1 cache 1;

create table "modules_names" (
	"module_id" integer not null default nextval('id'::text) primary key,
	"name" varchar(128)
);

create table "modules" (
	"module_id" integer,
	"ip" inet not null default '127.0.0.1',
	"port" integer,
	"keepalive" integer not null default 300,
	"ping_time_max" integer not null default 50,
	"crypted" integer default 0,
	"ziplink" integer default 0,
		foreign key ("module_id") references "modules_names" on delete cascade on update cascade
);

create table "modules_communication" (
	"module_id" integer,
	"request" varchar(128),
	"answer_to" integer,
	"answer" varchar(128),
	"timeout" integer,
	"order_no" integer,
		foreign key ("module_id") references "modules_names" on delete cascade on update cascade,
		foreign key ("answer_to") references "modules_names" on delete cascade on update cascade
);

create table "billing_commands" (
	"timestamp" timestamp,
	"command" character varying(128)
);

create table "syslog" (
	"timestamp" timestamp,
	"loglevel" integer not null default 1,
	"message" character varying(200)
);

create table "account_hosts_names" (
	"host_id" integer not null default nextval('id'::text) primary key,
	"name" character varying(128) not null default 'localhost'
);

create table "account_hosts" (
	"host_id" integer,
	"type" character varying(32) not null default '',
	    foreign key ("host_id") references "account_hosts_names" on delete cascade on update cascade
);

create table "management_hosts_names" (
	"host_id" integer not null default nextval('id'::text) primary key,
	"name" character varying(128) not null default 'localhost'
);

create table "management_hosts" (
	"host_id" integer,
	"type" character varying(32),
	    foreign key ("host_id") references "management_hosts_names" on delete cascade on update cascade
);

create table "customers" (
	"cust_id" integer not null default nextval('id'::text) primary key,
	"system" boolean not null default 'false',
	"name" character varying(40) not null default '',
	"sname" character varying(40) not null default '',
	"phone1" character varying(20) not null default '',
	"phone2" character varying(20) not null default '',
	"icq" character varying(24) not null default '',
	"email" character varying(64) not null default '',
	"address" character varying(128) not null default '',
	"address2" character varying(128) not null default '',
	"address3" character varying(128) not null default '',
	"comments" character varying(300) not null default ''
);

create table "cash" (
        "cust_id" integer,
	"cash" double precision not null default 0,
	"credit" double precision not null default 0,
		foreign key ("cust_id") references "customers" on delete cascade on update cascade
);

create table "tariffs_names" (
	"tariff_id" integer not null default nextval('id'::text) primary key,
	"name" character varying(128)
);

create table "nets_names" (
	"net_id" integer not null default nextval('id'::text) primary key,
	"name" character varying(128)
);

create table "nets" (
	"net_id" integer not null,
	"cidr" cidr,
	    foreign key ("net_id") references "nets_names" on delete cascade on update cascade
);

create table "plans_names" (
	"plan_id" integer not null default nextval('id'::text) primary key,
	"name" character varying(128)
);

create table "logins" (
	"login_id" integer not null primary key default nextval('id'::text),
	"cust_id" integer not null,
	"login" character varying(32) not null unique,
	"acl0" integer default 0,
	"acl1" integer default 0,
	"acl2" integer default 0,
	"acl3" integer default 0,
	"paused" boolean default 'f',
	"paused_at" timestamp,
	"max_logins" integer default 1,
);

create table "ip_users" (
	"login_id" integer not null,
	"plan_id" integer not null default 0,
	"detailed" boolean default 'f',
	"ip_addr" integer not null default '0',
	"fixed_ip" cidr not null default '127.0.0.1',
	"mac" character varying(20) default '',
	"lan" integer default 0,
	"email" integer default 0,
	"hosting" integer default 0,
	"vpn" integer default 0,
	"dialup" integer default 0,
	"ftp" integer default 0,
	"shell" integer default 0,
	"a_host_id" integer not null default 0,
	"m_host_id" integer not null default 0,
		foreign key ("cust_id") references "customers" on delete cascade on update cascade,
		foreign key ("plan_id") references "plans_names" on delete set default on update cascade,
		foreign key ("ip_addr") references "nets_names" ("net_id") on delete set default on update cascade,		
		foreign key ("a_host_id") references "account_hosts_names" ("host_id") on delete set default on update cascade,
		foreign key ("m_host_id") references "management_hosts_names" ("host_id") on delete set default on update cascade
);

create table "pwds_types" (
	"pwd_id" integer not null default nextval('id'::text) primary key,
	"type" varchar(16) not null default ''
);

create table "pwds_names" (
	"pwd_id" integer not null default 0,
	"name" character varying(64) not null,
	"crypt" character varying(64) not null default 'crypt',
	    foreign key ("pwd_id") references "pwds_types" on delete cascade on update cascade
);

create table "pwds" (
	"pwd_id" integer not null,
	"login_id" integer not null,
	"pwd" character varying(32) not null,
	    foreign key ("pwd_id") references "pwds_types" on delete cascade on update cascade,
	    foreign key ("login_id") references "logins" on delete cascade ON update cascade
);

create table "session_log" (
	"login_id" integer not null,
	"logged_from" character varying(32) not null default '',
	"ip" inet not null default '127.0.0.1',
	"login_time" timestamp not null default 'epoch',
	"logoff_time" timestamp not null default 'epoch',
	    foreign key ("login_id") references "logins" on delete cascade on update cascade
);

create index "session_id" on "session_log" ("login_id");
create index "session_login" on "session_log" ("login_time");
create index "session_logout" on "session_log" ("logoff_time");

create table "payment_history" (
	"cust_id" integer not null,
	"payment_date" timestamp not null default 'epoch',
	"amount" double precision default 0,
	"reason" character varying(60) not null default '',
	"registered_by" integer not null default 0,
	    foreign key ("cust_id") references "customers" on delete cascade on update cascade,
	    foreign key ("registered_by") references "logins" ("login_id") on delete set default on update set default
);

create table "services_names" (
	"service_id" integer not null default nextval('id'::text) primary key,
	"name" character varying(64)
);

create table "services" (
        "service_id" integer,
        "direction" character varying(8),
        "proto" character varying(16) not null default 'all',
        "net_id" integer,
        "sport" character varying(16) not null default '',
        "dport" character varying(16) not null default '',
        "iface" character varying(16) not null default '',
	"globality" integer not null default 1000,
	"a_host_id" integer not null default 0,
	"a_host_ext" character varying(128),
	"m_host_id" integer not null default 0,
	"m_host_ext" character varying(128),
	    foreign key ("service_id") references  "services_names" on delete cascade on update cascade,
	    foreign key ("net_id") references "nets_names" on delete cascade on update cascade,
	    foreign key ("a_host_id") references "account_hosts_names" ("host_id") on delete cascade on update cascade,
	    foreign key ("m_host_id") references "management_hosts_names" ("host_id") on delete cascade on update cascade
);

create table "traffic_log" (
	"login_id" integer not null,
	"service_id" integer not null,
	"bytes" bigint default 0 not null,
	"pkts" bigint default 0  not null,
	"timestamp" int4 not null default 0,
	    foreign key ("login_id") references "logins" on delete cascade on update cascade,
	    foreign key ("service_id") references "services_names" on delete cascade on update cascade
);

create index "traffic_log_id" on "traffic_log" ("login_id");
create index "traffic_log_time" on "traffic_log" ("timestamp");
create index "traffic_log_login_time" on "traffic_log" ("login_id", "timestamp");

-- generic periodical dates
-- generic date "usual days" with start=1 and end=365\6 is a must have
--
create table "dates_names" (
	"date_id" integer not null default nextval('id'::text) primary key,
	"name" character varying(128)
);

-- dates for the current year
--
create table "dates" (
	"date_id" integer not null default 0,
	"s_date" date not null default 'epoch',
	"e_date" date not null
);

-- periodical everyear dates
--
create table "periodical_dates" (
	"date_id" integer not null default 0,
	"start" date not null default 'epoch',
	"end" date not null default '1970-12-31',
	    foreign key ("date_id") references "dates_names" on delete cascade on update cascade
);

create table "prices_names" (
	"price_id" integer not null default nextval('id'::text) primary key,
	"name" character varying(150) not null
);

create table "prices" (
	"price_id" integer not null,
	"s_volume" bigint not null default 0,
	"e_volume" bigint not null default 999999999999,
	"s_time_of_day" time not null default '00:00:00',
	"e_time_of_day" time not null default '23:59:59',
	"s_day_of_week" int2 not null default 0,
	"e_day_of_week" int2 not null default 6,
	"date_id" integer not null,
	"price" double precision default 0,
	    foreign key ("price_id") references "prices_names" on delete cascade on update cascade,
	    foreign key ("date_id") references "dates_names" on delete cascade on update cascade
);

create table "plans" (
	"plan_id" integer not null default 0,
	"tariff_id" integer not null default 0,
	"allow_credit" boolean default 'true',
	    foreign key ("plan_id") references "plans_names" on delete cascade on update cascade,
	    foreign key ("tariff_id") references "tariffs_names" on delete cascade on update cascade
);

create table "payment_targets"
	"pay_for" integer not null default nextval('id'::text) primary key,
	"pay_string" varchar(48);
);

create table "tariffs" (
	"tariff_id" integer not null,
	"pay_for" integer not null default 0,
	"length" character varying(32) not null,
	"price_id" integer not null default 0,
	"priority" integer not null default 1000,
	"service_id" integer not null,
	    foreign key ("tariff_id") references "tariffs_names" on delete cascade on update cascade,
	    foreign key ("service_id") references "services_names" on delete cascade on update cascade,
	    foreign key ("price_id") references "prices_names" on delete set default on update cascade,
	    foreign key ("pay_for") references "payment_targets" on delete cascade on update cascade
);

create table "periods" (
	"period_id" integer not null default nextval('id'::text) primary key,
	"login_id" integer not null,
	"tariff_id" integer not null default 0,
	"p_start" timestamp not null default 'epoch',
	"p_end" timestamp not null default 'epoch',
	"volume" bigint not null default 0,
	"paid" double precision not null default 0,
	"open" boolean not null default 'true',
	    foreign key ("login_id") references "logins" on delete cascade on update cascade,
	    foreign key ("tariff_id") references "tariffs_names" on delete cascade on update cascade
);

create index "periods_login" on "periods" ("login_id");
create index "periods_start" on "periods" ("p_start");
create index "periods_end" on "periods" ("p_end");

create table "invoices" (
	"cust_id" integer not null primary key,
	"period_id" integer not null default 0,
	    foreign key ("cust_id") references "customers" on delete cascade on update cascade,
	    foreign key ("period_id") references "periods" on delete set default on update cascade
);

create table "billed" (
	"billed_on" timestamp not null default 'epoch'
);


-- functions goes here

-- get volume of used units
-- input: period_id
-- output: volume
--
create function "get_used_volume" (integer)
returns bigint
as 'select volume from periods where period_id = $1;'
language 'sql';

-- test, if today is date from "dates"
-- input: date_id
-- output: true/false
create function "is_today_date" (integer)
returns bigint
as 'select count(*)
    from 
	dates
    where
	date_id = $1 and
	s_date <= current_date and
	e_date >= current_date;'
language 'sql';

-- returns most exact date_id (having narrowest length) for today
-- input: none
-- output: date_id
--
create function "most_exact_date" ()
returns integer
as 'select date_id
    from 
	dates
    where
	s_date <= current_date and
	e_date >= current_date
    order by (e_date-s_date) asc limit 1;'
language 'sql';

-- extracts price for for one accounting unit 
-- input: period_id
-- output: price
--
create function "get_price" (integer)
returns double precision
as 'select price 
    from 
	prices, tariffs, periods 
    where 
	periods.tariff_id = tariffs.tariff_id and
	prices.price_id = tariffs.price_id and
	s_volume <= get_used_volume($1) and
	e_volume >= get_used_volume($1) and
	s_time_of_day <= cast (current_time(0) as time) and
	e_time_of_day >= cast (current_time(0) as time) and
	s_day_of_week <= extract(dow from current_date) and
	e_day_of_week >= extract(dow from current_date) and
	date_id = most_exact_date();'
language 'sql';

-- checks whether login_id has money
-- to open new period for tariff_id
-- input: login_id, tariff_id
-- output: boolean
--
create or replace function "has_money" (integer, integer) 
returns boolean as '
declare tmp double precision;
begin
select into tmp "credit" + "cash" from customers, logins
where
	logins.login_id = $1 and
	customers.cust_id = logins.cust_id;
if tmp<0 then
	return false;
end if;
select into tmp customers.cust_id from customers, plans, logins
where 
	logins.login_id = $1 and
	plans.tariff_id = $2 and
	customers.cust_id = logins.cust_id and
	logins.plan_id = plans.plan_id and
	(( plans.allow_credit and customers.cash <= 0 ) or
	(plans.allow_credit = false and customers.cash > 0));
if tmp is null then
	return false;
end if;
return true; 
end;'
language 'plpgsql';

-- some procedures goes here
--

-- fill dates table with data from periodical_dates
-- input: none
-- output: true
create or replace function "fill_in_dates" ()
returns boolean as '
declare tmp integer;
begin
	select into tmp most_exact_date();
	if tmp is null then 
		insert into dates (date_id, s_date, e_date)
			select date_id, date(extract(year from current_date()) || \'-\' || 
						extract(month from "start") || \'-\' ||
						extract(day from "start")),
					date(extract(year from current_date()) || \'-\' ||
						extract(month from "end") || \'-\' ||
						extract(day from "end"))
			from periodical_dates;
		return true;
	end if;
	return true;
end;'
language 'plpgsql';

-- check whether login_id uses tariff_id now
-- input: (login_id, tariff_is)
-- output: boolean
--
create or replace function "uses_tariff" (integer, integer)
returns boolean as '
declare tmp integer;
begin
	select into tmp login_id from plans, tariffs, logins
	where
		logins.login_id = $1 and
		tariffs.tariff_id = $2 and
		logins.plan_id = plans.plan_id and
		plans.tariff_id = $2;
	if tmp is null then 
		return false;
	end if;
	return true;
end;'
language 'plpgsql';

-- checks whether login_id already have period for tariff_id
-- always returns false if login_id is paused
-- input: login_id, tariff_id
-- output: true for no period and false in case there is one
--
create or replace function "no_open_period" (integer, integer)
returns boolean as '
declare tmp integer;
begin
	select into tmp logins.login_id from plans, logins, customers, tariffs, periods
	where
		logins.login_id = $1 and
		periods.login_id = $1 and
		tariffs.tariff_id = $2 and
		customers.cust_id = logins.cust_id and
		logins.plan_id = plans.plan_id and
		tariffs.tariff_id = plans.tariff_id and
		periods.tariff_id = tariffs.tariff_id and
		logins.pause = false and
		periods.p_end > current_timestamp(0);
	if tmp is null then
		return true;
	end if;
	return false;
end;'
language 'plpgsql';

-- procedure that closes old periods and creates new ones
-- for logins, that has enought cash
--
create or replace function "close_old_create_new_periods" ()
returns boolean as '
declare 
	new_time timestamp;
	tmp interval;
	clp record;
begin
	new_time := current_timestamp(0);
	for clp in select period_id, login_id, tariff_id, p_end 
			from periods where "p_end" <=now() loop
		if uses_tariff(clp.login_id, clp.tariff_id) and 
				no_open_period(clp.login_id, clp.tariff_id) then
			select into tmp length from tariffs where tariff_id = clp.tariff_id limit 1;
				insert into periods (login_id, tariff_id, p_start, p_end) values
					(clp.login_id, clp.tariff_id, clp.p_end,
							clp.p_end + tmp);
		end if;
	end loop;
	for clp in select login_id, tariffs.tariff_id from logins, plans, tariffs where
			logins.plan_id = plans.plan_id and
			plans.tariff_id = tariffs.tariff_id and
			no_open_period(logins.login_id, tariffs.tariff_id) loop
		select into tmp length from tariffs where tariff_id = clp.tariff_id limit 1;
		insert into periods (login_id, tariff_id, p_start, p_end) values 
			(clp.login_id, clp.tariff_id, current_timestamp(0), current_timestamp(0) + tmp);
	end loop;
	update periods set "open"=false where "open"=true and p_end <= new_time;
	return true;
end;'
language 'plpgsql';

-- some rules\triggers goes here

-- trigger on delete cidr from net
-- issues command to billing
--
create or replace function del_cidr_net () returns trigger as '
declare tmp1 varchar;
	tmp2 varchar;
begin
	tmp1 := old.cidr;
	tmp2 := old.net_id;
	insert into billing_commands ("timestamp", "command") values (current_timestamp(0), 
		\'net delete cidr \' || tmp1 || \' \' || tmp2);
	return null;
end;'
language 'plpgsql';
--
create trigger del_cidr_net after delete on nets
	for each row execute procedure del_cidr_net();


-- trigger on insert new cidr to some net
--
create or replace function ins_cidr_net () returns trigger as '
declare tmp1 varchar;
	tmp2 varchar;
begin
	tmp1 := new.cidr;
	tmp2 := new.net_id;
	insert into billing_commands ("timestamp", "command") values (current_timestamp(0), 
		\'net insert cidr \' || tmp1 || \' \' || tmp2);
	return null;
end;'
language 'plpgsql';
--
create trigger ins_cidr_net after insert on nets
	for each row execute procedure ins_cidr_net();

-- update trigger
--
create or replace function upd_cidr_net () returns trigger as '
declare
	oldcidr varchar;
	newcidr varchar;
	oldnet varchar;
	newnet varchar;
begin
	oldcidr := old.cidr; newcidr := new.cidr;
	oldnet := old.net_id; newnet := new.net_id;
	insert into billing_commands ("timestamp", "command") values (current_timestamp(0),
		\'net delete cidr \' || oldcidr || \' \' || oldnet);
	insert into billing_commands ("timestamp", "command") values (current_timestamp(0),
                \'net insert cidr \' || newcidr || \' \' || newnet);
	return null;
end;'
language 'plpgsql';
--
create trigger upd_cidr_net after update on nets
	for each row execute procedure upd_cidr_net();



-- some triggers for nets_names table
--
create or replace function ins_net_name () returns trigger as '
declare
	net_id varchar;
	name varchar;
	global varchar;
begin
	net_id := new.net_id; name := new.name; global := new.globality;
	insert into billing_commands ("timestamp", "command") values (current_timestamp(0),
		\'net create \' || net_id || \' \' || name || \' \' || global);
	return null;
end;'
language 'plpgsql';
--
create trigger ins_net_name after insert on nets_names
	for each row execute procedure ins_net_name();

-- delete network
--
create or replace function del_net_name () returns trigger as '
declare
        net_id varchar;
        name varchar;
        global varchar;
begin
        net_id := old.net_id; name := old.name; global := old.globality;
        insert into billing_commands ("timestamp", "command") values (current_timestamp(0),
                \'net destroy \' || net_id || \' \' || name || \' \' || global);
        return null;
end;'
language 'plpgsql';
--
create trigger del_net_name after delete on nets_names
        for each row execute procedure del_net_name();

-- update trigger

create or replace function upd_net_name () returns trigger as '
declare
	oldnet varchar;
	oldname varchr;
	oldglob varchar;
	newnet varchar;
	newname varchar;
	newglob varchar;
begin
	oldnet := old.net_id; oldname := old.name; oldglob := old.globality;
	newnet := new.net_id; newname := new.name; newglob := new.globality;
	insert into billing_commands ("timestamp", "command") values (current_timestamp(0),
		\'net destroy \' || oldnet || \' \' || oldname || \' \' || oldglob);
        insert into billing_commands ("timestamp", "command") values (current_timestamp(0),
                \'net create \' || newnet || \' \' || newname || \' \' || newglob);
end;'
language 'plpgsql';
--
create trigger upd_net_name after update on nets_names
	for each row execute procedure upd_net_name();
----


-- function, that have to be run at billing startup to clean out stale sessions and 
-- other cleanup
-- now we make sure that all sessions are properly(?) closed
--
create or replace function start_billing () returns boolean as '
declare
	ses_time_e timestamp;
begin
	select into ses_time_e billed_on from billed;
	update session_log set logoff_time = ses_time_e where logoff_time = \'epoch\';
	return true;
end;'
language 'plpgsql';

